/* queue.c */

/* VIS-5D version 3.0 */

/*
vis5d program for visualizing five dimensional gridded data sets
Copyright (C) 1990, 1991, 1992  Bill Hibbard, Brian Paul, Dave Santek,
and Andre Battaiola.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/



#include "globals.h"
#include "queue.h"
#include "sync.h"
#ifdef SINGLE_TASK
#  include "work.h"
#endif



#define QSIZE 3000


struct entry { int type;         /* the type of entry */
	       int i1, i2, i3;   /* integer arguments */
	       float f1, f2, f3, f4, f5;   /* float arguments */
             };


static struct entry queue[QSIZE];
static int qsize;
static int qhead, qtail;      /* remove from head, add to tail */

static LOCK qlock;
static SEMAPHORE qnotfull;
static SEMAPHORE qnotempty;




void init_queue()
{
   int i, j;

   ALLOC_LOCK( qlock );
/*   ALLOC_SEM( qnotfull, 1 );*/
   ALLOC_SEM( qnotempty, 0 );
   qsize = 0;
   qhead = qtail = 0;
}



void terminate_queue()
{
   FREE_LOCK( qlock );
/*   FREE_SEM( qnotfull );*/
   FREE_SEM( qnotempty );
}



int get_qsize()
{
   return qsize;
}



void add_qentry( urgent, type, i1, i2, i3, f1, f2, f3, f4, f5 )
int urgent, type, i1, i2, i3;
float f1, f2, f3, f4, f5;
{
   int pos, i, found=0;

   while (qsize==QSIZE-2) {
      if (Debug)
	 printf("QUEUE FULL!!!\n");
      sleep(1);  /* just wait a second */
   }

   LOCK_ON( qlock );

   /* check if already in the queue */
   pos = qhead;
   found = 0;

   for (i=0;i<qsize;i++) {
      if (queue[pos].type==type &&     /* only need to test */
	  queue[pos].i1==i1 &&         /* the first three fields */
	  queue[pos].i2==i2) {
	 /* already in queue, cancel it if urgent */
	 found = 1;
	 if (urgent)
	   queue[pos].type = TASK_NULL;
	 break;
      }
      else {
	 pos++;
	 if (pos==QSIZE)
	   pos = 0;
      }
   }

   if (urgent) {
      /* insert at head */
      if (qhead==0)
	 qhead = QSIZE-1;
      else
	 qhead--;
      pos = qhead;
      qsize++;
      SIGNAL_SEM( qnotempty );
   }
   else if (!found) {
      /* insert at tail */
      pos = qtail;
      qtail++;
      if (qtail==QSIZE)
	qtail = 0;
      qsize++;
      SIGNAL_SEM( qnotempty );
   }

   queue[pos].type = type;
   queue[pos].i1 = i1;
   queue[pos].i2 = i2;
   queue[pos].i3 = i3;
   queue[pos].f1 = f1;
   queue[pos].f2 = f2;
   queue[pos].f3 = f3;
   queue[pos].f4 = f4;
   queue[pos].f5 = f5;

   if (Debug) {
      if (urgent)
	printf("**URGENT** **URGENT** **URGENT** **URGENT** ");
      printf("ADDED AT POS=%d\n", pos );
   }

   LOCK_OFF( qlock );
}



void add_qentry2( urgent, type, i1, i2, i3, f1, f2, f3, f4, f5 )
int urgent, type, i1, i2, i3;
float f1, f2, f3, f4, f5;
{
   int pos, i, found=0;

   while (qsize==QSIZE-2) {
      printf("QUEUE FULL!!!\n");
   }

   LOCK_ON( qlock );

   if (urgent) {
      /* insert at head */
      if (qhead==0)
	 qhead = QSIZE-1;
      else
	 qhead--;
      pos = qhead;
   }
   else {
      /* check if already in the queue */
      found = 0;
      pos = qhead;
      for (i=0;i<qsize;i++) {
	 if (queue[pos].type==type &&
	     queue[pos].i1==i1 &&
	     queue[pos].i2==i2 &&
	     /*	  queue[pos].i3==i3 &&
		  queue[pos].f1==f1 &&
		  queue[pos].f2==f2 &&
		  queue[pos].f3==f3 &&
		  queue[pos].f4==f4 && */
	          queue[pos].f5==f5) {
	    /* already in queue */
	    found = 1;
	    break;
	 }
	 else {
	    pos++;
	    if (pos==QSIZE)
	      pos = 0;
	 }
      }
      if (!found) {
	 /* not found */
	 pos = qtail;
	 qtail++;
	 if (qtail==QSIZE)
	   qtail = 0;
      }
   }

   if (!found) {
      queue[pos].type = type;
      queue[pos].i1 = i1;
      queue[pos].i2 = i2;
      queue[pos].i3 = i3;
      queue[pos].f1 = f1;
      queue[pos].f2 = f2;
      queue[pos].f3 = f3;
      queue[pos].f4 = f4;
      queue[pos].f5 = f5;

      if (Debug) {
	 if (urgent)
	   printf("**URGENT** **URGENT** **URGENT** **URGENT** ");
	 printf("ADDED AT POS=%d\n", pos );
      }

      qsize++;
      SIGNAL_SEM( qnotempty );
   }

   LOCK_OFF( qlock );

}



/*** get_qentry *******************************************************
   Return an entry from the work queue.  If the queue is empty, this
   fuction blocks.
**********************************************************************/
void get_qentry( type, i1, i2, i3, f1, f2, f3, f4, f5 )
int *type, *i1, *i2, *i3;
float *f1, *f2, *f3, *f4, *f5;
{
   if (Debug) printf("get_qentry\n");

   while (qsize==0)
     WAIT_SEM( qnotempty );

   LOCK_ON( qlock );

   if (qsize>0) {
      /* remove from head */
      *type = queue[qhead].type;
      *i1 = queue[qhead].i1;
      *i2 = queue[qhead].i2;
      *i3 = queue[qhead].i3;
      *f1 = queue[qhead].f1;
      *f2 = queue[qhead].f2;
      *f3 = queue[qhead].f3;
      *f4 = queue[qhead].f4;
      *f5 = queue[qhead].f5;

      if (Debug) printf("REMOVED FROM POS=%d\n", qhead );

      if (*type!=TASK_QUIT) {
	 qhead++;
	 if (qhead==QSIZE)
	   qhead = 0;

	 qsize--;
	 /*if (qsize==QSIZE-3)
	   SIGNAL_SEM( qnotfull );*/
      }

   }
   else {
      *type = TASK_NULL;
   }
   LOCK_OFF( qlock );
   if (Debug) printf("return\n");
}



/*** request_quit *****************************************************
   This is called when we want to exit the program.  It puts a message
   on the queue that all 'work' threads are to exit.
**********************************************************************/
void request_quit()
{
   add_qentry( 1, TASK_QUIT, 0,0,0, 0.0,0.0,0.0,0.0,0.0 );
}



static void new_isosurface( time, parm, urgent )
int time, parm, urgent;
{
   if (Debug)
     printf("new_isosurface( %d, %d, %d )\n", time, parm, urgent );
#ifdef SINGLE_TASK
   calc_surface( time, parm, IsoLevel[parm], -1 );
#else
   add_qentry( urgent, TASK_ISOSURFACE, time, parm, 0,
	       0.0, 0.0, 0.0, 0.0, 0.0 );
#endif
}



/*** request_isosurface ***********************************************
   This function is called when we want a new isosurface computed
   for all times steps.
   Input:  time - current time step which is put at head of queue.
           parm - which parameter to make isosurface of.
**********************************************************************/
void request_isosurface( time, parm )
int time, parm;
{
   int i;

   new_isosurface( time, parm, 1 );
   for (i=0;i<NumTimes;i++) {
      if (i!=time)
	 new_isosurface( i, parm, 0 );
   }

}



static void new_hslice( time, parm, urgent )
int time, parm, urgent;
{
   if (HSliceTable[parm][time].valid==0 ||
       HSliceTable[parm][time].level!=HSliceLevel[parm] ||
       HSliceTable[parm][time].interval!=HSliceInterval[parm] ) {
#ifdef SINGLE_TASK
      calc_hslice( time, parm, HSliceInterval[parm], HSliceLevel[parm], -1 );
#else
      add_qentry( urgent, TASK_HSLICE, time, parm, 0,
		 0.0, 0.0, 0.0, 0.0, 0.0 );
#endif
   }
}


/*** request_hslice ***************************************************
   This function is called when we want a new horizontal contour line
   slice computed for all times steps.
   Input:  time - current time step which is put at head of queue.
           parm - which parameter to make hslice of.
**********************************************************************/
void request_hslice( time, parm )
int time, parm;
{
   int i;

   new_hslice( time, parm, 1 );
   for (i=0;i<NumTimes;i++) {
      if (i!=time)
	new_hslice( i, parm, 0 );
   }	      
}



static void new_vslice( time, parm, urgent )
int time, parm, urgent;
{
   if (VSliceTable[parm][time].valid==0 ||
       VSliceTable[parm][time].r1!=VSliceR1[parm] ||
       VSliceTable[parm][time].c1!=VSliceC1[parm] ||
       VSliceTable[parm][time].r2!=VSliceR2[parm] ||
       VSliceTable[parm][time].c2!=VSliceC2[parm] ||
       VSliceTable[parm][time].interval!=VSliceInterval[parm] ) {
      if (Debug) printf("new_vslice( %d, %d, %d )\n", time, parm, urgent );
#ifdef SINGLE_TASK
      calc_vslice( time, parm, VSliceInterval[parm],
		  VSliceR1[parm], VSliceC1[parm],
		  VSliceR2[parm], VSliceC2[parm], -1 );
#else
      add_qentry( urgent, TASK_VSLICE, time, parm, 0,
		 0.0, 0.0, 0.0, 0.0, 0.0 );
#endif
   }
}



/*** request_vslice ***************************************************
   This function is called when we want a new vertical contour line
   slice computed for all times steps.
   Input:  time - current time step which is put at head of queue.
           parm - which parameter to make vslice of.
**********************************************************************/
void request_vslice( time, parm )
int time, parm;
{
   int i;

   new_vslice( time, parm, 1 );
   i = (time+1) % NumTimes;
   while (i!=time) {
      new_vslice( i, parm, 0 );
      i = (i+1) % NumTimes;
   }
/*
   for (i=0;i<NumTimes;i++) {
      if (i!=time)
	new_vslice( i, parm, 0 );
   }
*/
}



static void new_chslice( time, parm, urgent )
int time, parm, urgent;
{
   if (CHSliceTable[parm][time].valid==0 ||
       CHSliceTable[parm][time].level!=CHSliceLevel[parm] ) {
#ifdef SINGLE_TASK
      calc_chslice( time, parm, CHSliceLevel[parm], -1 );
#else
      add_qentry( urgent, TASK_CHSLICE, time, parm, 0,
		 0.0, 0.0, 0.0, 0.0, 0.0 );
#endif
   }
}



/*** request_chslice **************************************************
   This function is called when we want a new horizontal color line
   slice computed for all times steps.
   Input:  time - current time step which is put at head of queue.
           parm - which parameter to make hslice of.
**********************************************************************/
void request_chslice( time, parm )
int time, parm;
{
   int i;

   new_chslice( time, parm, 1 );
   for (i=0;i<NumTimes;i++) {
      if (i!=time)
	new_chslice( i, parm, 0 );
   }	      
}




static void new_cvslice( time, parm, urgent )
int time, parm, urgent;
{
   if (CVSliceTable[parm][time].valid==0 ||
       CVSliceTable[parm][time].r1!=CVSliceR1[parm] ||
       CVSliceTable[parm][time].c1!=CVSliceC1[parm] ||
       CVSliceTable[parm][time].r2!=CVSliceR2[parm] ||
       CVSliceTable[parm][time].c2!=CVSliceC2[parm] ) {
      if (Debug)
	printf("new_cvslice( %d, %d, %d )\n", time, parm, urgent );
#ifdef SINGLE_TASK
      calc_cvslice( time, parm,
		  CVSliceR1[parm], CVSliceC1[parm],
		  CVSliceR2[parm], CVSliceC2[parm], -1 );
#else
      add_qentry( urgent, TASK_CVSLICE, time, parm, 0,
		 0.0, 0.0, 0.0, 0.0, 0.0 );
#endif
   }
}



/*** request_cvslice **************************************************
   This function is called when we want a new vertical color
   slice computed for all times steps.
   Input:  time - current time step which is put at head of queue.
           parm - which parameter to make cvslice of.
**********************************************************************/
void request_cvslice( time, parm )
int time, parm;
{
   int i;

   new_cvslice( time, parm, 1 );
   for (i=0;i<NumTimes;i++) {
      if (i!=time)
	new_cvslice( i, parm, 0 );
   }	      
}




static void new_hwindslice( time, ws, urgent )
int time, ws, urgent;
{
   if (HWindTable[ws][time].valid==0 ||
       HWindTable[ws][time].level!=HWindLevel[ws] ||
       HWindTable[ws][time].scale!=HWindScale[ws] ||
       HWindTable[ws][time].density!=HWindDensity[ws] ) {
#ifdef SINGLE_TASK
      calc_hwindslice( time, ws, HWindLevel[ws],
		       HWindScale[ws], HWindDensity[ws], -1 );
#else
      add_qentry( urgent, TASK_HWIND, time, ws, 0,
		 0.0, 0.0, 0.0, 0.0, 0.0 );
#endif
   }
}



/*** request_hwindslice ***********************************************
   This function is called when we want a new horizontal wind slice
   computed for all times steps.
   Input:  time - current time step which is put at head of queue.
           ws - which wind slice to compute in [0..WINDSLICES-1].
**********************************************************************/
void request_hwindslice( time, ws )
int time, ws;
{
   int i;

   new_hwindslice( time, ws, 1 );
   for (i=0;i<NumTimes;i++) {
      if (i!=time)
	new_hwindslice( i, ws, 0 );
   }	      
}




static void new_vwindslice( time, ws, urgent )
int time, ws, urgent;
{
   if (VWindTable[ws][time].valid==0 ||
       VWindTable[ws][time].r1!=VWindR1[ws] ||
       VWindTable[ws][time].c1!=VWindC1[ws] ||
       VWindTable[ws][time].r2!=VWindR2[ws] ||
       VWindTable[ws][time].c2!=VWindC2[ws] ||
       VWindTable[ws][time].scale!=VWindScale[ws] ||
       VWindTable[ws][time].density!=VWindDensity[ws] ) {
#ifdef SINGLE_TASK
      calc_vwindslice( time, ws, VWindR1[ws], VWindC1[ws],
		       VWindR2[ws], VWindC2[ws],
		       VWindScale[ws], VWindDensity[ws], -1 );
#else
      add_qentry( urgent, TASK_VWIND, time, ws, 0,
		 0.0, 0.0, 0.0, 0.0, 0.0 );
#endif
   }
}



/*** request_vwindslice ***********************************************
   This function is called when we want a new vertical wind slice
   computed for all times steps.
   Input:  time - current time step which is put at head of queue.
           ws - which wind slice to compute in [0..WINDSLICES-1].
**********************************************************************/
void request_vwindslice( time, ws )
int time, ws;
{
   int i;

   new_vwindslice( time, ws, 1 );
   for (i=0;i<NumTimes;i++) {
      if (i!=time)
	new_vwindslice( i, ws, 0 );
   }	      
}




/*** request_traj *****************************************************
   This is called to compute a new trajectory.
   Input:  row, col, lev - start pos in [0..Nr-1],[0..Nc-1],[0..Nl-1].
           time - start time in [0..NumTimes-1].
           group - which traj group in [0..TRAJGROUPS-1].
           rib - non-zero means make ribbon instead of line traj.
           step - integration step (~100.0)
           len - trajectory length multiplier.
**********************************************************************/
void request_traj( row, col, lev, time, group, rib, step, len )
float row, col, lev;
int time, group,rib;
float step, len;
{
#ifdef SINGLE_TASK
   calc_traj( row, col, lev, time, group, rib, step, len );
#else
   add_qentry( 1, TASK_TRAJ,
	       time, group, rib,
	       row, col, lev, step, len );
#endif
}



